@@ -16,12 +16,9 @@ import (
 	"strings"
 	"time"
 
-	"github.com/mattn/go-xmpp"     // BSD-3-Clause
-	"github.com/pborman/getopt/v2" // BSD-3-Clause
-)
-
-const (
-	VERSION = "0.3.0"
+	"github.com/ProtonMail/gopenpgp/v2/crypto" // MIT License
+	"github.com/mattn/go-xmpp"                 // BSD-3-Clause
+	"github.com/pborman/getopt/v2"             // BSD-3-Clause
 )
 
 type configuration struct {
@@ -32,6 +29,23 @@ type configuration struct {
 	resource string
 }
 
+func rcvStanzas(client *xmpp.Client, iqc chan xmpp.IQ, msgc chan xmpp.Chat) {
+	for {
+		received, err := client.Recv()
+		if err != nil {
+			log.Println(err)
+		}
+		switch v := received.(type) {
+		case xmpp.Chat:
+			msgc <- v
+		case xmpp.IQ:
+			iqc <- v
+		case xmpp.Presence:
+		default:
+		}
+	}
+}
+
 func readMessage(messageFilePath string) (string, error) {
 	var (
 		output string
@@ -73,15 +87,21 @@ func readMessage(messageFilePath string) (string, error) {
 }
 
 func main() {
+	type recipientsType struct {
+		Jid       string
+		OxKeyRing *crypto.KeyRing
+	}
 
 	var (
 		err                                       error
 		message, user, server, password, resource string
+		oxPrivKey                                 *crypto.Key
+		recipients                                []recipientsType
 	)
 
 	// Define command line flags.
 	flagHelp := getopt.BoolLong("help", 0, "Show help.")
-	flagHttpUpload := getopt.StringLong("http-upload", 0, "", "Send a file via http-upload.")
+	flagHTTPUpload := getopt.StringLong("http-upload", 0, "", "Send a file via http-upload.")
 	flagDebug := getopt.BoolLong("debug", 'd', "Show debugging info.")
 	flagServer := getopt.StringLong("jserver", 'j', "", "XMPP server address.")
 	flagUser := getopt.StringLong("username", 'u', "", "Username for XMPP account.")
@@ -103,41 +123,56 @@ func main() {
 		"Minimal TLS version. 10 (TLSv1.0), 11 (TLSv1.1), 12 (TLSv1.2) or 13 (TLSv1.3).")
 	flagVersion := getopt.BoolLong("version", 0, "Show version information.")
 	flagMUCPassword := getopt.StringLong("muc-password", 0, "", "Password for password protected MUCs.")
+	flagOx := getopt.BoolLong("ox", 0, "Use \"OpenPGP for XMPP\" encryption (experimental).")
+	flagOxGenPrivKeyRSA := getopt.BoolLong("ox-genprivkey-rsa", 0,
+		"Generate a private OpenPGP key (RSA 4096 bit) for the given JID and publish the "+
+			"corresponding public key.")
+	flagOxGenPrivKeyX25519 := getopt.BoolLong("ox-genprivkey-x25519", 0,
+		"Generate a private OpenPGP key (x25519) for the given JID and publish the "+
+			"corresponding public key.")
+	flagOxPassphrase := getopt.StringLong("ox-passphrase", 0, "",
+		"Passphrase for locking and unlocking the private OpenPGP key.")
+	flagOxImportPrivKey := getopt.StringLong("ox-import-privkey", 0, "",
+		"Import an existing private OpenPGP key.")
+	flagOxDeleteNodes := getopt.BoolLong("ox-delete-nodes", 0, "Delete existing OpenPGP nodes on the server.")
 
 	// Parse command line flags.
 	getopt.Parse()
 
-	// If requested, show help and quit.
-	if *flagHelp {
+	switch {
+	case *flagHelp:
+		// If requested, show help and quit.
 		getopt.Usage()
 		os.Exit(0)
-	}
-
-	// If requested, show version and quit.
-	if *flagVersion {
-		fmt.Println("go-sendxmpp", VERSION)
+	case *flagVersion:
+		// If requested, show version and quit.
+		fmt.Println("go-sendxmpp", version)
 		fmt.Println("License: BSD-2-clause")
 		os.Exit(0)
+		// Quit if Ox (OpenPGP for XMPP) is requested for unsupported operations like
+		// groupchat, http-upload or listening.
+	case *flagOx && *flagHTTPUpload != "":
+		log.Fatal("No Ox support for http-upload available.")
+	case *flagOx && *flagChatroom:
+		log.Fatal("No Ox support for chat rooms available.")
+	case *flagHTTPUpload != "" && *flagInteractive:
+		log.Fatal("Interactive mode and http upload can't" +
+			" be used at the same time.")
+	case *flagHTTPUpload != "" && *flagMessageFile != "":
+		log.Fatal("You can't send a message while using" +
+			" http upload.")
 	}
 
 	// Read recipients from command line and quit if none are specified.
 	// For listening or sending raw XML it's not required to specify a recipient except
 	// when sending raw messages to MUCs (go-sendxmpp will join the MUC automatically).
-	recipients := getopt.Args()
-	if (len(recipients) == 0 && !*flagRaw && !*flagListen) ||
-		(len(recipients) == 0 && *flagChatroom) {
+	recipientsList := getopt.Args()
+	if (len(recipientsList) == 0 && !*flagRaw && !*flagListen && !*flagOxGenPrivKeyX25519 &&
+		!*flagOxGenPrivKeyRSA && *flagOxImportPrivKey == "") && !*flagOxDeleteNodes ||
+		(len(recipientsList) == 0 && *flagChatroom) {
 		log.Fatal("No recipient specified.")
 	}
 
-	// Check that all recipient JIDs are valid.
-	for i, recipient := range recipients {
-		validatedJid, err := MarshalJID(recipient)
-		if err != nil {
-			log.Fatal(err)
-		}
-		recipients[i] = validatedJid
-	}
-
 	// Read configuration file if user or password is not specified.
 	if *flagUser == "" || *flagPassword == "" {
 		// Read configuration from file.
@@ -175,18 +210,7 @@ func main() {
 		resource = *flagResource
 	} else if resource == "" {
 		// Use "go-sendxmpp" plus a random string if no other resource is specified
-		resource = "go-sendxmpp." + getID()
-	}
-
-	if (*flagHttpUpload != "") && (*flagInteractive || (*flagMessageFile != "")) {
-		if *flagInteractive {
-			log.Fatal("Interactive mode and http upload can't" +
-				" be used at the same time.")
-		}
-		if *flagMessageFile != "" {
-			log.Fatal("You can't send a message while using" +
-				" http upload.")
-		}
+		resource = "go-sendxmpp." + getShortID()
 	}
 
 	// Timeout
@@ -244,9 +268,88 @@ func main() {
 		log.Fatal(err)
 	}
 
-	if *flagHttpUpload != "" {
-		message = httpUpload(client, tlsConfig.ServerName,
-			*flagHttpUpload)
+	iqc := make(chan xmpp.IQ, 100)
+	msgc := make(chan xmpp.Chat, 100)
+	go rcvStanzas(client, iqc, msgc)
+
+	for _, r := range getopt.Args() {
+		var re recipientsType
+		re.Jid = r
+		if *flagOx {
+			re.OxKeyRing, err = oxGetPublicKeyRing(client, iqc, r)
+			if err != nil {
+				re.OxKeyRing = nil
+				fmt.Println("Couldn't receive key for:", r)
+			}
+		}
+		recipients = append(recipients, re)
+	}
+
+	// Check that all recipient JIDs are valid.
+	for i, recipient := range recipients {
+		validatedJid, err := MarshalJID(recipient.Jid)
+		if err != nil {
+			log.Fatal(err)
+		}
+		recipients[i].Jid = validatedJid
+	}
+
+	switch {
+	case *flagOxGenPrivKeyX25519:
+		validatedOwnJid, err := MarshalJID(user)
+		if err != nil {
+			log.Fatal(err)
+		}
+		err = oxGenPrivKey(validatedOwnJid, client, iqc, *flagOxPassphrase, "x25519")
+		if err != nil {
+			log.Fatal(err)
+		}
+		os.Exit(0)
+	case *flagOxGenPrivKeyRSA:
+		validatedOwnJid, err := MarshalJID(user)
+		if err != nil {
+			log.Fatal(err)
+		}
+		err = oxGenPrivKey(validatedOwnJid, client, iqc, *flagOxPassphrase, "rsa")
+		if err != nil {
+			log.Fatal(err)
+		}
+		os.Exit(0)
+	case *flagOxImportPrivKey != "":
+		validatedOwnJid, err := MarshalJID(user)
+		if err != nil {
+			log.Fatal(err)
+		}
+		err = oxImportPrivKey(validatedOwnJid, *flagOxImportPrivKey,
+			client, iqc)
+		if err != nil {
+			log.Fatal(err)
+		}
+		os.Exit(0)
+	case *flagOxDeleteNodes:
+		validatedOwnJid, err := MarshalJID(user)
+		if err != nil {
+			log.Fatal(err)
+		}
+		err = oxDeleteNodes(validatedOwnJid, client, iqc)
+		if err != nil {
+			log.Fatal(err)
+		}
+		os.Exit(0)
+	case *flagOx:
+		validatedOwnJid, err := MarshalJID(user)
+		if err != nil {
+			log.Fatal(err)
+		}
+		oxPrivKey, err = oxGetPrivKey(validatedOwnJid, *flagOxPassphrase)
+		if err != nil {
+			log.Fatal(err)
+		}
+	}
+
+	if *flagHTTPUpload != "" {
+		message = httpUpload(client, iqc, tlsConfig.ServerName,
+			*flagHTTPUpload)
 	}
 
 	// Skip reading message if '-i' or '--interactive' is set to work with e.g. 'tail -f'.
@@ -266,8 +369,6 @@ func main() {
 
 			if err := scanner.Err(); err != nil {
 				if err != io.EOF {
-					// Close connection and quit.
-					_ = client.Close()
 					log.Fatal(err)
 				}
 			}
@@ -279,186 +380,142 @@ func main() {
 	reg := regexp.MustCompile(`[\x{0000}-\x{0008}\x{000B}\x{000C}\x{000E}-\x{001F}]`)
 	message = reg.ReplaceAllString(message, "")
 
-	// Send raw XML to chatroom
-	if *flagChatroom && *flagRaw {
-		var err error
+	var msgType string
+	msgType = "chat"
+	if *flagChatroom {
+		msgType = "groupchat"
 		// Join the MUCs.
 		for _, recipient := range recipients {
 			if *flagMUCPassword != "" {
 				dummyTime := time.Now()
-				_, err = client.JoinProtectedMUC(recipient, *flagResource,
+				_, err = client.JoinProtectedMUC(recipient.Jid, *flagResource,
 					*flagMUCPassword, xmpp.NoHistory, 0, &dummyTime)
 			} else {
-				_, err = client.JoinMUCNoHistory(recipient, *flagResource)
+				_, err = client.JoinMUCNoHistory(recipient.Jid, *flagResource)
 			}
 			if err != nil {
-				// Try to nicely close connection,
-				// even if there was an error joining.
-				_ = client.Close()
 				log.Fatal(err)
 			}
 		}
-		// Send raw XMP
+	}
+	switch {
+	case *flagRaw:
+		// Send raw XML
 		_, err = client.SendOrg(message)
 		if err != nil {
-			// Try to nicely close connection,
-			// even if there was an error sending.
-			_ = client.Close()
 			log.Fatal(err)
 		}
-		// After sending the message, leave the MUCs
-		for _, recipient := range recipients {
-			_, err = client.LeaveMUC(recipient)
-			if err != nil {
-				log.Println(err)
+	case *flagInteractive:
+		// Send in endless loop (for usage with e.g. "tail -f").
+		for {
+			scanner := bufio.NewScanner(os.Stdin)
+			scanner.Scan()
+			message = scanner.Text()
+			for _, recipient := range recipients {
+				switch {
+				case *flagOx:
+					if recipient.OxKeyRing == nil {
+						continue
+					}
+					oxMessage, err := oxEncrypt(client, iqc, oxPrivKey,
+						recipient.Jid, recipient.OxKeyRing, message)
+					if err != nil {
+						fmt.Println("Ox: couldn't encrypt to",
+							recipient.Jid)
+						continue
+					}
+					_, err = client.SendOrg(oxMessage)
+					if err != nil {
+						log.Fatal(err)
+					}
+				default:
+					_, err = client.Send(xmpp.Chat{Remote: recipient.Jid,
+						Type: msgType, Text: message})
+					if err != nil {
+						log.Fatal(err)
+					}
+				}
 			}
 		}
-		// Wait for a short time as some messages are not delievered by the server
-		// if the connection is closed immediately after sending a message.
-		time.Sleep(100 * time.Millisecond)
-		return
-	}
-
-	if *flagListen {
+	case *flagListen:
+		tz := time.Now().Location()
 		for {
-			received, err := client.Recv()
-			if err != nil {
-				log.Println(err)
-			}
-
-			switch v := received.(type) {
-			case xmpp.Chat:
-				if v.Text == "" {
+			v := <-msgc
+			switch {
+			case isOxMsg(v) && *flagOx:
+				msg, t, err := oxDecrypt(v, client, iqc, user, oxPrivKey)
+				if err != nil {
+					log.Println(err)
+					continue
+				}
+				if msg == "" {
 					continue
 				}
-				t := time.Now()
 				bareFrom := strings.Split(v.Remote, "/")[0]
 				// Print any messages if no recipients are specified
 				if len(recipients) == 0 {
-					fmt.Println(t.Format(time.RFC3339), bareFrom+":", v.Text)
+					fmt.Println(t.In(tz).Format(time.RFC3339), "[OX]",
+						bareFrom+":", msg)
 				} else {
 					for _, recipient := range recipients {
-						if bareFrom == strings.ToLower(recipient) {
-							fmt.Println(t.Format(time.RFC3339), bareFrom+":", v.Text)
+						if bareFrom == strings.ToLower(recipient.Jid) {
+							fmt.Println(t.In(tz).Format(time.RFC3339),
+								"[OX]", bareFrom+":", msg)
 						}
 					}
 				}
 			default:
-				continue
-			}
-		}
-	}
-
-	// Send message to chatroom(s) if the flag is set.
-	if *flagChatroom {
-
-		for _, recipient := range recipients {
-
-			// Join the MUC.
-			if *flagMUCPassword != "" {
-				dummyTime := time.Now()
-				_, err = client.JoinProtectedMUC(recipient, *flagResource,
-					*flagMUCPassword, xmpp.NoHistory, 0, &dummyTime)
-			} else {
-				_, err = client.JoinMUCNoHistory(recipient, *flagResource)
-			}
-			if err != nil {
-				// Try to nicely close connection,
-				// even if there was an error joining.
-				_ = client.Close()
-				log.Fatal(err)
-			}
-		}
-
-		// Send in endless loop (for usage with e.g. "tail -f").
-		if *flagInteractive {
-			for {
-				scanner := bufio.NewScanner(os.Stdin)
-				scanner.Scan()
-				message = scanner.Text()
-				for _, recipient := range recipients {
-					_, err = client.Send(xmpp.Chat{Remote: recipient,
-						Type: "groupchat", Text: message})
-					if err != nil {
-						// Try to nicely close connection,
-						// even if there was an error sending.
-						_ = client.Close()
-						log.Fatal(err)
+				var t time.Time
+				if v.Text == "" {
+					continue
+				}
+				if v.Stamp.IsZero() {
+					t = time.Now()
+				} else {
+					t = v.Stamp
+				}
+				bareFrom := strings.Split(v.Remote, "/")[0]
+				// Print any messages if no recipients are specified
+				if len(recipients) == 0 {
+					fmt.Println(t.In(tz).Format(time.RFC3339), bareFrom+":", v.Text)
+				} else {
+					for _, recipient := range recipients {
+						if bareFrom == strings.ToLower(recipient.Jid) {
+							fmt.Println(t.In(tz).Format(time.RFC3339),
+								bareFrom+":", v.Text)
+						}
 					}
 				}
 			}
-		} else {
-			// Send the message.
-			for _, recipient := range recipients {
-				if *flagHttpUpload != "" {
-					_, err = client.Send(xmpp.Chat{Remote: recipient,
-						Type: "groupchat", Ooburl: message, Text: message})
-				} else {
-					_, err = client.Send(xmpp.Chat{Remote: recipient,
-						Type: "groupchat", Text: message})
+		}
+	default:
+		for _, recipient := range recipients {
+			switch {
+			case *flagHTTPUpload != "":
+				_, err = client.Send(xmpp.Chat{Remote: recipient.Jid,
+					Type: msgType, Ooburl: message, Text: message})
+				if err != nil {
+					fmt.Println("Couldn't send message to",
+						recipient.Jid)
 				}
+			case *flagOx:
+				if recipient.OxKeyRing == nil {
+					continue
+				}
+				oxMessage, err := oxEncrypt(client, iqc, oxPrivKey,
+					recipient.Jid, recipient.OxKeyRing, message)
+				if err != nil {
+					fmt.Println("Ox: couldn't encrypt to", recipient.Jid)
+					continue
+				}
+				_, err = client.SendOrg(oxMessage)
 				if err != nil {
-					// Try to nicely close connection,
-					// even if there was an error sending.
-					_ = client.Close()
 					log.Fatal(err)
 				}
-			}
-		}
-
-		for _, recipient := range recipients {
-			// After sending the message, leave the Muc
-			_, err = client.LeaveMUC(recipient)
-			if err != nil {
-				log.Println(err)
-			}
-		}
-	} else {
-		// Send raw XML
-		if *flagRaw {
-			_, err = client.SendOrg(message)
-			if err != nil {
-				// Try to nicely close connection,
-				// even if there was an error sending.
-				_ = client.Close()
-				log.Fatal(err)
-			}
-			// Wait for a short time as some messages are not delievered by the server
-			// if the connection is closed immediately after sending a message.
-			time.Sleep(100 * time.Millisecond)
-			return
-		}
-
-		// Send in endless loop (for usage with e.g. "tail -f").
-		if *flagInteractive {
-			for {
-				scanner := bufio.NewScanner(os.Stdin)
-				scanner.Scan()
-				message = scanner.Text()
-				for _, recipient := range recipients {
-					_, err = client.Send(xmpp.Chat{Remote: recipient,
-						Type: "chat", Text: message})
-					if err != nil {
-						// Try to nicely close connection,
-						// even if there was an error sending.
-						_ = client.Close()
-						log.Fatal(err)
-					}
-				}
-			}
-		} else {
-			for _, recipient := range recipients {
-				if *flagHttpUpload != "" {
-					_, err = client.Send(xmpp.Chat{Remote: recipient, Type: "chat",
-						Ooburl: message, Text: message})
-				} else {
-					_, err = client.Send(xmpp.Chat{Remote: recipient, Type: "chat",
-						Text: message})
-				}
+			default:
+				_, err = client.Send(xmpp.Chat{Remote: recipient.Jid,
+					Type: msgType, Text: message})
 				if err != nil {
-					// Try to nicely close connection,
-					// even if there was an error sending.
-					_ = client.Close()
 					log.Fatal(err)
 				}
 			}
@@ -468,10 +525,4 @@ func main() {
 	// Wait for a short time as some messages are not delievered by the server
 	// if the connection is closed immediately after sending a message.
 	time.Sleep(100 * time.Millisecond)
-
-	// Close XMPP connection
-	err = client.Close()
-	if err != nil {
-		log.Fatal(err)
-	}
 }
